/***********************************************************************
*
* asmdef.h - Assembler header for the TI 990 computer.
*
* Changes:
*   05/21/03   DGP   Original.
*	
***********************************************************************/

#include <stdio.h>

/*
** Definitions
*/

#define NORMAL 0
#define ABORT  12

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define VERSION "1.0.6"

#define COMMENTSYM '*'
#define PCSYM '$'
#define MEMSYM '@'
#define HEXSYM '>'
#define STRINGSYM '\''
#define EOFSYM ':'

#define MAXLINE 256
#define LINESPAGE 60
#define NOOPERAND 30
#define TTLSIZE 60
#define IDTSIZE 8

#define TEMPSPEC "asXXXXXX"

/*
** Object output formats
*/

#define OBJFORMAT "%c%04X"
#define IDTFORMAT "%c%04X%-8.8s"
#define REFFORMAT "%c%4.4X%-6.6s"
#define DEFFORMAT "%c%4.4X%-6.6s"
#define BYTEFORMAT "%c%02X00"
#define SEQFORMAT "%6.6d\n"

/*
** Listing formats
*/

#define H1FORMAT "ASM990 %-8.8s  %-24.24s   %s Page %04d\n"
#define H2FORMAT "%s\n\n"

#define L1FORMAT " %s  %s %4d %s"
#define L2FORMAT " %s  %s\n"

#define PCBLANKS "    "
#define PCFORMAT "%04X"

#define OPBLANKS  "                 "
#define OP1FORMAT "%04X             "
#define OP2FORMAT "%04X  %04X%c      "
#define OP3FORMAT "%04X  %04X%c %04X%c"
#define ADDRFORMAT "            %04X%c"
#define CHAR1FORMAT "%02X"
#define DATA1FORMAT "%04X%c"

#define SYMFORMAT " %-8.8s %4.4X%c "

/*
** Data type definitions
*/

#define int8            char
#define int16           short
#define int32           int
typedef int             t_stat;                         /* status */
typedef int             t_bool;                         /* boolean */
typedef unsigned int8   uint8;
typedef unsigned int16  uint16;
typedef unsigned int32  uint32, t_addr;                 /* address */

#if defined (WIN32)                                     /* Windows */
#define t_int64 __int64
#elif defined (__ALPHA) && defined (VMS)                /* Alpha VMS */
#define t_int64 __int64
#elif defined (__ALPHA) && defined (__unix__)           /* Alpha UNIX */
#define t_int64 long
#else                                                   /* default GCC */
#define t_int64 long long
#endif                                                  /* end OS's */
typedef unsigned t_int64        t_uint64, t_value;      /* value */
typedef t_int64                 t_svalue;               /* signed value */

#if defined(__s390__) || defined(OPENMVS) || defined(sparc)
#define MSL 0
#define LSL 1
#else
#define MSL 1
#define LSL 0
#endif

/*
** OpCode table definition
*/

#define NUMOPS 91

enum optypes
{
   TYPE_1=1,
   TYPE_2,
   TYPE_3,
   TYPE_4,
   TYPE_5,
   TYPE_6,
   TYPE_7,
   TYPE_8,
   TYPE_9,
   TYPE_P  	/* Pseudo op */
};

enum psops
{
   AORG_T=1,
   BES_T,
   BSS_T,
   BYTE_T,
   DATA_T,
   DEF_T,
   DORG_T,
   DXOP_T,
   END_T,
   EQU_T,
   EVEN_T,
   IDT_T,
   LIST_T,
   NOP_T,
   OPTION_T,
   PAGE_T,
   REF_T,
   RORG_T,
   RT_T,
   TEXT_T,
   TITL_T,
   UNL_T,
};

typedef struct
{
   char *opcode;
   unsigned opvalue;
   unsigned opmod;
   int optype;
} OpCode;

/*
** Symbol table
*/

#define MAXSYMLEN 6
#define MAXSYMBOLS 2000

typedef struct
{
   char symbol[32];
   int value;
   int relocatable;
   int external;
   int global;
   int xop;
} SymNode;

/*
** Token types from tokscan
*/

enum toktypes
{
   PC = 1,
   SYM,
   MEMREF,
   REG,
   DECNUM,
   HEXNUM,
   STRING,
   INVALID,
   EOL
};
 
/*
** Expression scanner states.
*/

enum states
{
   START_STATE,
   ADD_STATE,
   SUB_STATE,
   MUL_STATE,
   DIV_STATE,
   END_STATE
};

/*
** Object tags
*/

#define IDT_TAG		'0'
#define ABSENTRY_TAG	'1'
#define RELENTRY_TAG	'2'
#define RELEXTRN_TAG	'3'
#define ABSEXTRN_TAG	'4'
#define RELGLOBAL_TAG	'5'
#define ABSGLOBAL_TAG	'6'
#define CKSUM_TAG	'7'
#define NOCKSUM_TAG	'8'
#define ABSORG_TAG	'9'
#define RELORG_TAG	'A'
#define ABSDATA_TAG	'B'
#define RELDATA_TAG	'C'
#define EOR_TAG		'F'

#define CHARSPERREC	66  /* Chars per object record */
#define WORDTAGLEN	5   /* Word + Object Tag length */
#define EXTRNLEN	11  /* Tag + SYMBOL + addr */
#define GLOBALLEN	11  /* Tag + SYMBOL + addr */
#define IDTLEN		13  /* Tag + IDT + length */
#define SEQUENCENUM	74  /* Where to put the sequence */

/*
** Pass 1 error table - pass 2 does the printing.
*/

#define MAXERRORS 1000

typedef struct
{
   char *errortext;
   int errorline;
} ErrTable;

/*
** External definitions
*/

extern int asmpass1 (FILE *);
extern int asmpass2 (FILE *, FILE *, int, FILE *);

extern OpCode *oplookup (char *);

extern char *tokscan (char *, char **, int *, char *);
extern SymNode *symlookup (char *, int);
extern char *exprscan (char *, int *, char *, int *, int, int, int);

