/***********************************************************************
*
* asmpass1 - Pass 1 for the TI 990 assembler.
*
* Changes:
*   05/21/03   DGP   Original.
*   07/02/03   DGP   Change hanging byte methods.
*   07/10/03   DGP   Mark inital REF as absolute.
*   07/15/03   DGP   Adjust symbol value on an EVEN.
*	
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "asmdef.h"

extern int pc;
extern int rorgpc;
extern int symbolcount;
extern int absolute;
extern int indorg;
extern int errcount;
extern int p1errcnt;
extern int pgmlength;
extern char inbuf[MAXLINE];
extern SymNode *symbols[MAXSYMBOLS];
extern ErrTable p1error[MAXERRORS];
extern char ttlbuf[TTLSIZE+2];
extern char idtbuf[IDTSIZE+2];

static int linenum;
static char cursym[32];

static int bytecnt;

/***********************************************************************
* p11op - Type 1 processor.
***********************************************************************/

static void
p11op (OpCode *op, char *bp)
{
   char *tok;
   int toktyp;
   int val;
   char term;

   while (isspace (*bp)) bp++;
   if (*bp == MEMSYM)
      pc += 2;
   while (*bp && *bp != ',') bp++;
   if (*bp == ',')
   {
      bp++;
      if (*bp == MEMSYM)
	 pc += 2;
   }
   pc += 2;
}

/***********************************************************************
* p12op - Type 2 processor.
***********************************************************************/

static void
p12op (OpCode *op, char *bp)
{
   pc += 2;
}

/***********************************************************************
* p13op - Type 3 processor.
***********************************************************************/

static void
p13op (OpCode *op, char *bp)
{
   char *tok;
   int toktyp;
   int val;
   char term;

   while (isspace (*bp)) bp++;
   if (*bp == MEMSYM)
      pc += 2;
   pc += 2;
}

/***********************************************************************
* p14op - Type 4 processor.
***********************************************************************/

static void
p14op (OpCode *op, char *bp)
{
   char *tok;
   int toktyp;
   int val;
   char term;

   while (isspace (*bp)) bp++;
   if (*bp == MEMSYM)
      pc += 2;
   pc += 2;
}

/***********************************************************************
* p15op - Type 5 processor.
***********************************************************************/

static void
p15op (OpCode *op, char *bp)
{
   pc += 2;
}

/***********************************************************************
* p16op - Type 6 processor.
***********************************************************************/

static void
p16op (OpCode *op, char *bp)
{
   char *tok;
   int toktyp;
   int val;
   char term;

   while (isspace (*bp)) bp++;
   if (*bp == MEMSYM)
      pc += 2;
   pc += 2;
}

/***********************************************************************
* p17op - Type 7 processor.
***********************************************************************/

static void
p17op (OpCode *op, char *bp)
{
   pc += 2;
}

/***********************************************************************
* p18op - Type 8 processor.
***********************************************************************/

static void
p18op (OpCode *op, char *bp)
{
   if (op->opmod != 2) pc += 2;
   pc += 2;
}

/***********************************************************************
* p19op - Type 9 processor.
***********************************************************************/

static void
p19op (OpCode *op, char *bp)
{
   char *tok;
   int toktyp;
   int val;
   char term;

   while (isspace (*bp)) bp++;
   if (*bp == MEMSYM)
      pc += 2;
   pc += 2;
}

/***********************************************************************
* p1pop - Process Pseudo ops.
***********************************************************************/


static void
p1pop (OpCode *op, char *bp)
{
   SymNode *s;
   char *cp;
   char *tok;
   int toktyp;
   int relocatable;
   int val;
   char term;

   switch (op->opvalue)
   {

   case AORG_T:
      if (!absolute && !indorg) rorgpc = pc;
      cp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      if (val < 0 || val > 65535)
      {
	 val = 0;
      }
      absolute = TRUE;
      indorg = FALSE;
      pc = val;
      if (cursym[0])
      {
	 s = symlookup (cursym, FALSE);
	 s->value = val;
	 s->relocatable = FALSE;
      }
      break;

   case BES_T:
      bp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      if (val < 0 || val > 65535)
      {
	 val = 0;
      }
      pc += val;
      if (cursym[0])
      {
	 s = symlookup (cursym, FALSE);
	 s->value = pc;
	 s->relocatable = relocatable;
      }
      break;

   case BSS_T:
      bp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      if (val < 0 || val > 65535)
      {
	 val = 0;
      }
      pc += val;
      break;

   case BYTE_T:
      do {
	 bp = exprscan (bp, &val, &term, &relocatable, 2, TRUE, 0);
	 bytecnt++;
	 pc++;
      } while (term == ',');
      break;

   case DATA_T:
      if (pc & 0x0001) pc++;
#ifdef DEBUGPC
      printf ("DATA pc = %04X\n", pc);
#endif
      if (cursym[0])
      {
	 s = symlookup (cursym, FALSE);
	 s->value = pc;
      }
      do {
	 bp = exprscan (bp, &val, &term, &relocatable, 2, TRUE, 0);
	 pc += 2;
#ifdef DEBUGPC
	 printf ("    pc = %04X\n", pc);
#endif
      } while (term == ',');
      break;

   case DEF_T:
      break;

   case DORG_T:
      if (!absolute && !indorg) rorgpc = pc;
      cp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      if (val < 0 || val > 65535)
      {
	 val = 0;
      }
      absolute = TRUE;
      indorg = TRUE;
      pc = val;
      if (cursym[0])
      {
	 s = symlookup (cursym, FALSE);
	 s->value = val;
	 s->relocatable = FALSE;
      }
      break;

   case DXOP_T:
      bp = tokscan (bp, &tok, &toktyp, &term);
      if (toktyp == SYM)
      {
	 s = symlookup (tok, TRUE);
	 if (term == ',')
	 {
	    bp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
	    if (val < 0 || val > 15)
	    {
	       errcount++;
	       p1error[p1errcnt].errorline = linenum;
	       p1error[p1errcnt].errortext = (char *)malloc (120);
	       sprintf (p1error[p1errcnt].errortext, "Invalid DXOP value: %d",
			val);
	       p1errcnt++;
	       val = 0;
	    }
	    s->value = val;
	    s->relocatable = FALSE;
	    s->xop = TRUE;
	 }
      }
      else
      {
	 errcount++;
	 p1error[p1errcnt].errorline = linenum;
	 p1error[p1errcnt].errortext = (char *)malloc (120);
	 sprintf (p1error[p1errcnt].errortext, "Invalid DXOP symbol: %s",
		  tok);
	 p1errcnt++;
      }
      break;

   case END_T:
      pgmlength = pc;
      break;

   case EQU_T:
      cp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
#ifdef DEBUGPC
      printf ("EQU pc = %04X, val = %04X\n", pc, val);
#endif
      if (cursym[0])
      {
	 s = symlookup (cursym, FALSE);
	 s->value = val;
	 s->relocatable = relocatable;
      }
      break;

   case EVEN_T:
      if (pc & 0x0001) pc++;
      if (cursym[0])
      {
	 s = symlookup (cursym, FALSE);
	 s->value = pc;
	 s->relocatable = relocatable;
      }
      break;

   case IDT_T:
      if (!idtbuf[0])
      {
	 while (isspace(*bp)) bp++;
	 if (*bp == '\'')
	 {
	    cp = ++bp;
	    while (*bp != '\'' && *bp != '\n') bp++;
	 }
	 *bp = '\0';
	 if (strlen(idtbuf) > IDTSIZE)
	 {
	    *(cp+IDTSIZE) = '\0';
	 }
	 strcpy (idtbuf, cp);
      }
      break;

   case NOP_T:
   case RT_T:
      if (pc & 0x0001) pc++;
      pc += 2;
      break;

   case OPTION_T:
      break;

   case PAGE_T:
      break;

   case REF_T:
      do {
	 bp = tokscan (bp, &tok, &toktyp, &term);
	 if (toktyp != SYM)
	 {
	    errcount++;
	    p1error[p1errcnt].errorline = linenum;
	    p1error[p1errcnt].errortext = (char *)malloc (120);
	    sprintf (p1error[p1errcnt].errortext, "REF requires symbol: %s",
		     tok);
	    p1errcnt++;
	    return;
	 }
	 if (strlen(tok) > MAXSYMLEN)
	 {
	    tok[MAXSYMLEN] = '\0';
	 }
	 if ((s = symlookup (tok, FALSE)) == NULL)
	 {
	    s = symlookup (tok, TRUE);
	 }
	 s->value = 0;
	 s->relocatable = FALSE;
	 s->external = TRUE;
      } while (term == ',');
      break;

   case RORG_T:
      bp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      indorg = FALSE;
      absolute = FALSE;
      if (val == 0)
	 pc = rorgpc;
      else
         pc = val;
      if (cursym[0])
      {
	 s = symlookup (cursym, FALSE);
	 s->value = pc;
	 s->relocatable = TRUE;
      }
      break;

   case TEXT_T:
      bp = tokscan (bp, &tok, &toktyp, &term);
      if (term == '-')
	 bp = tokscan (bp, &tok, &toktyp, &term);
      if (toktyp == STRING)
      {
	 bytecnt = strlen(tok);
	 pc += bytecnt;
      }
      break;

   case TITL_T:
      if (!ttlbuf[0])
      {
	 while (isspace(*bp)) bp++;
	 if (*bp == '\'')
	 {
	    cp = ++bp;
	    while (*bp != '\'' && *bp != '\n') bp++;
	 }
	 *bp = '\0';
	 strcpy (ttlbuf, cp);
      }
      break;

   case UNL_T:
      break;

   default: ;
   }
}

/***********************************************************************
* asmpass1 - Pass 1
***********************************************************************/

int
asmpass1 (FILE *tmpfd)
{
   char *tok;
   int status = 0;
   int toktyp;
   char term;

#ifdef DEBUG
   printf ("asmpass1: Entered\n");
#endif

   /*
   ** Rewind the input.
   */

   if (fseek (tmpfd, 0, SEEK_SET) < 0)
   {
      perror ("asm990: Can't rewind temp file");
      return (-1);
   }

   /*
   ** Process the source.
   */

   pc = 0;
   rorgpc = 0;

   absolute = FALSE;
   indorg = FALSE;

   linenum = 0;
   ttlbuf[0] = '\0';
   idtbuf[0] = '\0';

   while (fgets(inbuf, MAXLINE, tmpfd))
   {
      char *bp;
      char *cp;

#ifdef DEBUG
      printf ("in = %s", inbuf);
#endif
      linenum++;
      bp = inbuf;

      /*
      ** If not a comment, then process.
      */

      if (*bp != COMMENTSYM)
      {
	 SymNode *s;
	 OpCode *op;

	 /*
	 ** If label present, add to symbol table.
	 */

         if (isalnum(*bp))
	 {
	    cp = tokscan (bp, &tok, &toktyp, &term);
	    strcpy (cursym, tok);
	    if (strlen(tok) > MAXSYMLEN)
	    {
	       cursym[MAXSYMLEN] = '\0';
	       tok[MAXSYMLEN] = '\0';
	    }
	    if ((s = symlookup (tok, TRUE)) == NULL)
	    {
	       errcount++;
	       p1error[p1errcnt].errorline = linenum;
	       p1error[p1errcnt].errortext = (char *)malloc (120);
	       sprintf (p1error[p1errcnt].errortext, "Duplicate symbol: %s",
		        tok);
	       p1errcnt++;
	       status = -1;
	    }
	    bp = cp;
	    if (term == '\n') continue;
	 }
	 else 
	 {
	    cursym[0] = '\0';
	    while (isspace (*bp)) bp++;
	 }

	 /*
	 ** Scan off opcode.
	 */

	 cp = tokscan (bp, &tok, &toktyp, &term);
	 if (!tok[0]) continue;
#ifdef DEBUGPC
	 printf ("INST = %s, pc = %04X\n", tok, pc);
#endif

	 if ((op = oplookup (tok)) != NULL)
	 {
	    bytecnt = 0;
	    bp = cp;
	    if ((op->optype != TYPE_P) && (pc & 0x0001)) pc++;
	    switch (op->optype)
	    {

	       case TYPE_1:
	          p11op (op, bp);
		  break; 

	       case TYPE_2:
	          p12op (op, bp);
		  break; 

	       case TYPE_3:
	          p13op (op, bp);
		  break; 

	       case TYPE_4:
	          p14op (op, bp);
		  break; 

	       case TYPE_5:
	          p15op (op, bp);
		  break; 

	       case TYPE_6:
	          p16op (op, bp);
		  break; 

	       case TYPE_7:
	          p17op (op, bp);
		  break; 

	       case TYPE_8:
	          p18op (op, bp);
		  break; 

	       case TYPE_9:
	          p19op (op, bp);
		  break; 

	       case TYPE_P:
	          p1pop (op, bp);
	          break;

	       default: ;
	    }
	 }
	 else
	 {
	    /* print errors in pass2 */
	    pc++; 
	    status = -1;
	 }
	 
      }

   }

   return (status);
}
