/***********************************************************************
*
* asmsupt.c - Support routines for the TI 990 assembler.
*
* Changes:
*   05/21/03   DGP   Original.
*   07/10/03   DGP   If external is absolute and val 0, change relocatable.
*	
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "asmdef.h"

extern int pc;
extern int symbolcount;
extern int errcount;
extern int errnum;
extern int absolute;
extern int pgmlength;
extern char inbuf[MAXLINE];
extern SymNode *symbols[MAXSYMBOLS];
extern char errline[10][120];


/***********************************************************************
* tokscan - Scan a token.
***********************************************************************/

char *
tokscan (char *buf, char **tok, int *toktyp, char *term)
{
   char *cp;
   int tt = 0;
   static char t[80];
   char c;

#ifdef DEBUGTOK
   printf ("tokscan: Entered: buf = %s", buf);
#endif
   while (isspace(*buf))
   {
      /*
      ** We stop scanning on a newline or when we're passed the operands.
      */

      if (*buf == '\n' || (buf - inbuf > NOOPERAND))
      {
	 t[0] = '\0';
	 *tok = t;
	 *term = '\n';
	 *toktyp = EOL;
         return (buf);
      }
      buf++;
   }

   cp = buf;
   if (*buf == PCSYM) 
   {
      tt = PC;
      t[0] = *buf++;
   }
   else if (*buf == MEMSYM)
   {
      t[0] = '\0';
      *tok = t;
      *toktyp = MEMREF;
      *term = *buf;
      buf++;
      return (buf);
   }
   else if (*buf == STRINGSYM)
   {
      tt = STRING;
      buf++;
      cp = buf;
   }
   else if (*buf == HEXSYM)
   {
      tt = HEXNUM;
      buf++;
      cp = buf;
   }
   else if (isalpha(*buf)) tt = SYM;
   else if (isdigit(*buf)) tt = DECNUM;

   if (tt == STRING)
   {
      char *sp;
      char string[80];
      sp = string;
      while (*buf) 
      {
         if (*buf != '\'')
	    *sp++ = *buf++;
         else if (*(buf+1) == '\'')
	 {
	    *sp++ = *buf++; buf++;
	 }
	 else break;
      }
      buf++;
      *sp++ = '\0';
      cp = string;
   }
   else
   {
      while (*buf
	    && *buf != '+' && *buf != '-' && *buf != ',' && *buf != '*'
	    && *buf != '/' && *buf != '(' && *buf != ')' && *buf != '\'')
      {
	 if (isspace (*buf)) break;
	 if (tt == DECNUM && !isdigit(*buf)) tt = INVALID;
	 if (tt == HEXNUM && !isxdigit(*buf)) tt = INVALID;
	 buf++;
      }
   }
   c = *buf;
   *buf = '\0';
   if (strlen (cp) > (sizeof(t)-1))
      *(cp+sizeof(t)-1) = '\0';
   strcpy (t, cp);
   if (c == '\n')
      *buf = c;
   else
      *buf++ = c;
   *tok = t;
   *toktyp = tt;
   *term = c;
#ifdef DEBUGTOK
   printf ("  tok = %s, toktyp = %d, term = %02x\n", t, tt, c);
#endif
   return (buf);
}

/***********************************************************************
* symlookup - Lookup a symbol in the symbol table.
***********************************************************************/

SymNode *
symlookup (char *sym, int add)
{
   SymNode *ret = NULL;
   int done = FALSE;
   int mid;
   int last = 0;
   int lo;
   int up;
   int r;

#ifdef DEBUGSYM
   printf ("symlookup: Entered: sym = %s\n", sym);
#endif

   /*
   ** Empty symbol table.
   */

   if (symbolcount == 0)
   {
      if (!add) return (NULL);

#ifdef DEBUGSYM
      printf ("add symbol at top\n");
#endif
      if ((symbols[symbolcount] = (SymNode *)malloc (sizeof (SymNode))) == NULL)
      {
         fprintf (stderr, "asm990: Unable to allocate memory\n");
	 exit (ABORT);
      }
      strcpy (symbols[0]->symbol, sym);
      symbols[0]->value = pc;
      symbols[0]->external = FALSE;
      symbols[0]->global = FALSE;
      symbols[0]->xop = FALSE;
      if (absolute)
	 symbols[0]->relocatable = FALSE;
      else
	 symbols[0]->relocatable = TRUE;
      symbolcount++;
      return (symbols[0]);
   }

   /*
   ** Locate using binary search
   */

   lo = 0;
   up = symbolcount;
   last = -1;
   
   while (!done)
   {
      mid = (up - lo) / 2 + lo;
#ifdef DEBUGSYM
      printf (" mid = %d, last = %d\n", mid, last);
#endif
      if (last == mid) break;
      r = strcmp (symbols[mid]->symbol, sym);
      if (r == 0)
      {
	 if (add) return (NULL); /* must be a duplicate */
         return (symbols[mid]);
      }
      else if (r < 0)
      {
         lo = mid;
      }
      else 
      {
         up = mid;
      }
      last = mid;
   }

   /*
   ** Not found, check to add
   */

   if (add)
   {
      SymNode *new;

#ifdef DEBUGSYM
      printf ("add new symbol\n");
#endif
      if (symbolcount+1 > MAXSYMBOLS)
      {
         fprintf (stderr, "asm990: Symbol table exceeded\n");
	 exit (ABORT);
      }

      if ((new = (SymNode *)malloc (sizeof (SymNode))) == NULL)
      {
         fprintf (stderr, "asm990: Unable to allocate memory\n");
	 exit (ABORT);
      }

      strcpy (new->symbol, sym);
      new->value = pc;
      new->external = FALSE;
      new->global = FALSE;
      new->xop = FALSE;
      if (absolute)
	 new->relocatable = FALSE;
      else
	 new->relocatable = TRUE;

      /*
      ** Insert pointer in sort order.
      */

      for (lo = 0; lo < symbolcount; lo++)
      {
         if (strcmp (symbols[lo]->symbol, sym) > 0)
	 {
	    for (up = symbolcount + 1; up > lo; up--)
	    {
	       symbols[up] = symbols[up-1];
	    }
	    symbols[lo] = new;
	    symbolcount++;
	    return (symbols[lo]);
	 }
      }
      symbols[symbolcount] = new;
      ret = symbols[symbolcount];
      symbolcount++;
   }
   return (ret);
}

/***********************************************************************
* exprscan - Scan an expression and return its value.
***********************************************************************/

char *
exprscan (char *bp, int *val, char *tterm, int *relo, int entsize, int defer,
	  int pcinc)
{
   SymNode *s;
   char *cp;
   char *tok;
   int tokval;
   int relocatable = 0;
   int external = FALSE;
   int toktyp;
   int lval = 0;
   int state = START_STATE;
   char term;

#ifdef DEBUGEXPR
   printf ("exprscan: Entered: expr = %s\n", bp);
#endif

   while (state != END_STATE)
   {
      cp = tokscan (bp, &tok, &toktyp, &term);
#ifdef DEBUGEXPR
      printf (" state = %d, tok = %s, toktyp = %d, term = %2x\n",
	 state, tok, toktyp, term);
#endif
      switch (toktyp)
      {

      case MEMREF:
         break;

      case SYM:
	 if (!defer)
	 {
	    if (strlen(tok) > MAXSYMLEN)
	    {
	       tok[MAXSYMLEN] = '\0';
	    }
	    if ((s = symlookup (tok, FALSE)) == NULL)
	    {
	       errcount++;
	       sprintf (errline[errnum++], "Undefined symbol: %s", tok);
	       tokval = 0;
	    }
	    else
	    {
	       /* Mark REF as being used */
	       if (s->external && s->value == 0)
		  s->relocatable = !absolute;
	       if (s->relocatable) relocatable++;
	       tokval = s->value;
	       if (s->external)
	       {
		  external = TRUE;
		  /* Don't relocate since it's external tail */
		  if (tokval == 0) relocatable = 100;
		  /* back link reference chain */
		  s->value = pc + pcinc; 
	       }
	    }
	 }
	 else /* defer look now */
	    tokval = 0;
	 break;

      case PC:
	 relocatable++;
	 tokval = pc;
	 break;

      case DECNUM:
	 tokval = strtol (tok, NULL, 10);
	 break;

      case HEXNUM:
	 tokval = strtol (tok, NULL, 16);
	 break;

      case STRING:
         if (strlen(tok) > entsize)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "String too long for expression: %s",
		     tok);
	    tokval = 0;
	 }
	 else
	 {
	    if (entsize == 1 || tok[1] == '\0')
	       tokval = tok[0] & 0xFF;
	    else /* 990 only allows 1 & 2 */
	       tokval = tok[0] << 8 | tok[1];
	 }
	 break;

      default:
	 tokval = 0;
	 break;
      }

      bp = cp;
#ifdef DEBUGEXPR
      printf (" tokval = %d(%X)\n", tokval, tokval);
#endif

      switch (state)
      {

      case START_STATE:
	 lval = tokval;
	 break;

      case ADD_STATE:
	 if (external)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid external reference: %s", tok);
	    lval = 0;
	 }
	 else
	    lval += tokval;
	 break;

      case SUB_STATE:
	 if (external)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid external reference: %s", tok);
	    lval = 0;
	 }
	 else
	    lval -= tokval;
	 break;

      case MUL_STATE:
	 if (external)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid external reference: %s", tok);
	    lval = 0;
	 }
	 else
	 {
	    relocatable = 100;
	    lval *= tokval;
	 }
	 break;

      case DIV_STATE:
	 if (external)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid external reference: %s", tok);
	    lval = 0;
	 }
	 else
	 {
	    relocatable = 100;
	    lval /= tokval;
	 }
	 break;

      default: ;
      }

#ifdef DEBUGEXPR
      printf (" lval = %d(%X)\n", lval, lval);
#endif
      switch (term)
      {

      case '+':
         state = ADD_STATE;
	 break;

      case '-':
         state = SUB_STATE;
	 break;

      case '*':
         state = MUL_STATE;
	 break;

      case '/':
         state = DIV_STATE;
	 break;

      default:
         state = END_STATE;
      }

   }

#ifdef DEBUGEXPR
   printf (" value = %d(%X), relocatable = %d\n", lval, lval, relocatable);
#endif

   if (!absolute && relocatable == 1) *relo = TRUE;
   else *relo = FALSE;
   *val = lval;
   *tterm = term;
   return (bp);
}
